from __future__ import annotations

import dataclasses
import enum
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from .action import Action
    from .project import Project

AnyTask: TypeAlias = "Task[Action | None]"
T_Action = TypeVar("T_Action", bound="Action | None", covariant=True)


class TaskCaptureMode(enum.Enum):
    """Describes how the output of the task should be captured, if at all."""

    #: The task output should not be captured.
    NONE = enum.auto()

    #: The task output should be captured, but always be shown in the build.
    SEMI = enum.auto()

    #: The task output should be fully captured and not shown (unless the task fails).
    FULL = enum.auto()


@dataclasses.dataclass
class Task(Generic[T_Action]):
    """Represents a logical unit of work."""

    name: str
    project: Project
    action: T_Action
    metadata: list[Any] = dataclasses.field(default_factory=list)
    dependencies: list[AnyTask] = dataclasses.field(default_factory=list)  #: Strict dependencies
    after: list[AnyTask] = dataclasses.field(default_factory=list)  #: Optional dependencies
    before: list[AnyTask] = dataclasses.field(default_factory=list)  #: Optional predecessors
    default: bool = True
    capture: TaskCaptureMode = TaskCaptureMode.FULL

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r})"

    @property
    def path(self) -> str:
        if self.project.parent is None:
            return f":{self.name}"
        else:
            return f"{self.project.path}:{self.name}"

    def finalize(self) -> None:
        """This method is called by :meth:`BuildContext.finalize()`. It gives the task a chance update its
        configuration before the build process is executed. Most commonly, custom task implementations will
        initialize their :attr:`action`, as the delayed action creation is often the main reason to creating
        a Task subclass in the first place."""



from __future__ import annotations

import abc
import enum


class ActionResult(enum.Enum):
    UP_TO_DATE = enum.auto()
    SKIPPED = enum.auto()
    SUCCEEDED = enum.auto()
    FAILED = enum.auto()


class Action(abc.ABC):
    """Actions implement the behaviour of tasks."""

    def is_up_to_date(self) -> bool:
        """Gives the action a chance before it is executed to inform the build executor that it is up
        to date and does not need to be executed. Some actions may be able to determine this quickly so
        they can implement this method to improve build performance and user information display."""

        return False

    def is_skippable(self) -> bool:
        """Gives the action a chance before it is executed to inform the build executor that the action
        can be skipped. This status is different from :meth:`is_up_to_date` but may lead to the same result,
        i.e. that the action is not executed."""

        return False

    @abc.abstractmethod
    def execute(self) -> ActionResult:
        ...



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Iterable, Iterator, Optional, Type, TypeVar, Union, cast

from kraken.core.task import TaskCaptureMode
from kraken.core.utils import flatten

if TYPE_CHECKING:
    from .action import Action
    from .build_context import BuildContext
    from .task import AnyTask, Task

ProjectMember = Union["Project", "Task"]
T_ProjectMember = TypeVar("T_ProjectMember", bound=ProjectMember)
T_Task = TypeVar("T_Task", bound="AnyTask")
T_Action = TypeVar("T_Action", bound="Action")


class Project:
    """A project consolidates related tasks."""

    directory: Path
    name: str
    parent: Optional[Project]
    context: BuildContext

    def __init__(self, directory: Path, name: str, parent: Optional[Project], context: BuildContext) -> None:
        self.directory = directory
        self.name = name
        self.parent = parent
        self.context = context

        # We store all members that can be referenced by a fully qualified name in the same
        # dictionary to ensure we're not accidentally allocating the same name twice.
        self._members: dict[str, ProjectMember] = {}

    def __repr__(self) -> str:
        return f'Project(path="{self.path}")'

    @property
    def path(self) -> str:
        if self.parent is None:
            return ":"
        elif self.parent.parent is None:
            return f":{self.name}"
        else:
            return f"{self.parent.path}:{self.name}"

    @property
    def members(self) -> ProjectMembers[ProjectMember]:
        """Returns the superset of all members in the project."""

        return ProjectMembers(self, self._members)

    @property
    def tasks(self) -> ProjectTasks:
        return ProjectTasks(self, self._members)

    @property
    def children(self) -> ProjectChildren:
        return ProjectChildren(self, self._members)

    def do(
        self,
        name: str,
        action: T_Action,
        default: bool = True,
        capture: TaskCaptureMode = TaskCaptureMode.FULL,
        dependencies: Iterable[AnyTask | str] = (),
        after: Iterable[AnyTask | str] = (),
        before: Iterable[AnyTask | str] = (),
    ) -> Task[T_Action]:
        """Add a task to the project under the given name, executing the specified action."""

        from .task import Task

        def _resolve_tasks(tasks: Iterable[AnyTask | str]) -> list[AnyTask]:
            return list(
                flatten(self.context.resolve_tasks([task]) if isinstance(task, str) else [task] for task in tasks)
            )

        task = Task(
            name,
            self,
            action,
            dependencies=_resolve_tasks(dependencies),
            after=_resolve_tasks(after),
            before=_resolve_tasks(before),
            default=default,
            capture=capture,
        )
        self.tasks.add(task)
        return task


class ProjectMembers(Generic[T_ProjectMember]):
    """Container for the members of a project."""

    def __init__(
        self,
        project: "Project",
        members: dict[str, ProjectMember],
    ) -> None:
        self._project = project
        self._members = members

    @staticmethod
    def _type() -> Optional[Type[T_ProjectMember]]:
        return None

    def __len__(self) -> int:
        _type = self._type()
        if _type is None:
            return len(self._members)
        # TODO (@niklas.rosenstein): Imperformant; maybe split tasks/projects into separate dicts?
        return sum(1 for _ in self)

    def __contains__(self, value: object) -> bool:
        _type = self._type()
        if value in self._members:
            if _type is not None:
                member = self._members[cast(str, value)]
                return isinstance(member, _type)
            return True
        return False

    def __iter__(self) -> Iterator[T_ProjectMember]:
        _type = self._type()
        for value in self._members.values():
            if _type is None or isinstance(value, _type):
                yield cast(T_ProjectMember, value)

    def __getitem__(self, name: str) -> T_ProjectMember:
        try:
            member = self._members[name]
        except KeyError:
            raise ValueError(f"{self._project} has no member {name!r}")
        _type = self._type()
        if _type is not None and not isinstance(member, _type):
            raise ValueError(
                f"{self._project} member {name!r} is a {type(member).__name__} " f"but not a {_type.__name__}"
            )
        return cast(T_ProjectMember, member)

    def keys(self) -> set[str]:
        _type = self._type()
        if _type is None:
            return cast(set[str], self._members.keys())
        else:
            result = set()
            for key, value in self._members.items():
                if _type is None or isinstance(value, _type):
                    result.add(key)
            return result


class ProjectTasks(ProjectMembers["AnyTask"]):
    @staticmethod
    def _type() -> Optional[Type[Any]]:
        from .task import Task

        return Task

    def add(self, task: Task[T_Action]) -> Task[T_Action]:
        if task.project is not self._project:
            raise RuntimeError("Task.project does not match")
        if task.name in self._members:
            raise ValueError(f"{self._project} already has a member {task.name!r}")
        self._members[task.name] = task
        return task


class ProjectChildren(ProjectMembers["Project"]):
    @staticmethod
    def _type() -> Optional[Type[Project]]:
        return Project

    def add(self, project: Project) -> Project:
        if project.parent is not self._project:
            raise RuntimeError("Project.parent does not match")
        if project.name in self._members:
            raise ValueError(f"{self._project} already has a member {project.name!r}")
        self._members[project.name] = project
        return project



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Callable

from kraken.core.action import Action, ActionResult


@dataclasses.dataclass
class RenderFileAction(Action):
    content: str | Callable[[], str]
    file: Path
    encoding: str = "utf-8"

    def is_up_to_date(self) -> bool:
        if callable(self.content):
            self.content = self.content()
        return self.file.is_file() and self.file.read_text() == self.content

    def execute(self) -> ActionResult:
        if callable(self.content):
            self.content = self.content()
        self.file.parent.mkdir(exist_ok=True)
        encoded = self.content.encode(self.encoding)
        print(f"write {self.file} ({len(encoded)} bytes)")
        self.file.write_bytes(encoded)
        return ActionResult.SUCCEEDED



""" Implements a loader for Python scripts. These Python scripts have access to importing certain build context
related members from the :mod:`kraken.api` module; note that these members are only available when imported in
the context of a script executed by this loader and not otherwise. """

from __future__ import annotations

import sys
import types
from pathlib import Path
from typing import Any

from kraken.core.project import Project

from . import BuildScriptLoader

DEFAULT_FILE = Path("kraken.py")


class PyscriptLoader(BuildScriptLoader):
    def detect_in_project_directory(self, project_dir: Path) -> Path | None:
        file = project_dir / DEFAULT_FILE
        return file if file.is_file() else None

    def match_file(self, file: Path) -> bool | Path | None:
        return file.suffix == ".py"

    def load_script(self, file: Path, project: Project) -> None:
        api_module_name = "kraken.api"

        # Create the temporary replacement for the kraken.api module that the script will import from.
        api_module: Any = types.ModuleType(api_module_name)
        api_module.ctx = project.context
        api_module.project = project

        # In order for @dataclass decorators to work in a Python script loaded by this build, it must be
        # able to look up the module in sys.modules.
        module = types.ModuleType(f"_kraken__{project.name}_{id(project)}")
        module.__file__ = str(file)

        old_module = sys.modules.get(api_module_name)
        try:
            sys.modules[api_module_name] = api_module
            sys.modules[module.__name__] = module
            code = compile(file.read_text(), filename=file, mode="exec")
            exec(code, vars(module))
        finally:
            del sys.modules[module.__name__]
            if old_module is None:
                sys.modules.pop(api_module_name)
            else:
                sys.modules[api_module_name] = old_module



from __future__ import annotations

from typing import Iterable, TypeVar

T = TypeVar("T")


def flatten(it: Iterable[Iterable[T]]) -> Iterable[T]:
    for item in it:
        yield from item


def not_none(v: T | None) -> T:
    if v is None:
        raise RuntimeError("expected not-None")
    return v



from __future__ import annotations

import dataclasses
from typing import Iterable, List, cast

from networkx import DiGraph  # type: ignore[import]

from kraken.core.task import AnyTask
from kraken.core.utils import not_none


@dataclasses.dataclass
class _Node:
    task: AnyTask
    required: bool


@dataclasses.dataclass
class _Edge:
    strict: bool


class BuildGraph:
    """Represents the build graph."""

    def __init__(self, tasks: Iterable[AnyTask] = ()) -> None:
        """Create a new build graph from the given task list."""

        # Nodes have the form {'data': _Node} and edges have the form {'data': _Edge}.
        self._digraph = DiGraph()
        self._add_tasks(tasks)

    def __bool__(self) -> bool:
        return len(self._digraph.nodes) > 0

    # Low level internal API

    def _get_node(self, task_path: str) -> _Node | None:
        data = self._digraph.nodes.get(task_path)
        if data is None:
            return None
        return cast(_Node, data["data"])

    def _add_node(self, task: AnyTask, required: bool) -> None:
        """Internal. Ensures that a node for the given task exists. If *required* is True, it will override the
        existing status."""

        node = self._get_node(task.path) or _Node(task, required)
        node.required = node.required or required
        self._digraph.add_node(task.path, data=node)

    def _get_edge(self, task_a: str, task_b: str) -> _Edge | None:
        data = self._digraph.edges.get((task_a, task_b))
        if data is None:
            return None
        return cast(_Edge, data["data"])

    def _add_edge(self, task_a: str, task_b: str, strict: bool) -> None:
        """Internal. Introduces an edge between two tasks."""

        edge = self._get_edge(task_a, task_b) or _Edge(strict)
        edge.strict = edge.strict or strict
        self._digraph.add_edge(task_a, task_b, data=edge)

    # High level internal API

    def _add_task(
        self,
        task: AnyTask,
        strict_dependencies: Iterable[AnyTask],
        optional_dependencies: Iterable[AnyTask],
        required: bool,
    ) -> None:
        """Internal. Adds the given task and it's strict and optional dependencies to the graph."""

        self._add_node(task, required)

        for dependency in strict_dependencies:
            self._add_node(dependency, False)
            self._add_edge(dependency.path, task.path, True)

        for dependency in optional_dependencies:
            self._add_node(dependency, False)
            self._add_edge(dependency.path, task.path, False)

    def _add_tasks(self, tasks: Iterable[AnyTask], required: bool = True) -> None:
        """Internal. Extends the internal directed graph by the given tasks.

        Args:
            tasks: The tasks to add to the graph.
            required: Whether the tasks added via this function should be marked as required (i.e. they
                will never be ignored by the :meth:`trimmed` method).
        """

        for task in tasks:
            self._add_task(task, task.dependencies, task.after, required)
            for dependant in task.before:
                self._add_task(dependant, (), [task], False)

    # Public API

    def get_predecessors(self, task: AnyTask) -> List[AnyTask]:
        return [not_none(self._get_node(task_path)).task for task_path in self._digraph.predecessors(task.path)]

    def get_successors(self, task: AnyTask) -> List[AnyTask]:
        return [not_none(self._get_node(task_path)).task for task_path in self._digraph.successors(task.path)]

    def trim(self) -> BuildGraph:
        """Removes all tasks from the graph that are not initially required and only connected to any other
        task with an optional dependency."""

        # Find all tasks that are not initially required and only have dependants with an optional dependency.
        weakly_connected_tasks = set()
        for task_path in self._digraph.nodes:
            if not_none(self._get_node(task_path)).required:
                continue
            if not any(
                not_none(self._get_edge(task_path, dependant_task_path)).strict
                for dependant_task_path in self._digraph.successors(task_path)
            ):
                weakly_connected_tasks.add(task_path)

        # Remove the subgraphs that are weakly connected.
        def _remove_subgraph(task_path: str) -> None:
            if task_path not in self._digraph.nodes:
                return
            for successor in list(self._digraph.predecessors(task_path)):
                _remove_subgraph(successor)
            self._digraph.remove_node(task_path)

        for task_path in weakly_connected_tasks:
            if task_path in self._digraph.nodes:
                _remove_subgraph(task_path)

        return self

    def execution_order(self) -> Iterable[AnyTask]:
        from networkx.algorithms import topological_sort  # type: ignore[import]

        order = topological_sort(self._digraph)
        return (not_none(self._get_node(task_path)).task for task_path in order)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src', 'api': 'src/api'}

packages = \
['api', 'core', 'core.action', 'core.loaders']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.6.0,<3.0.0', 'setuptools>=33.1.0']

entry_points = \
{'kraken.core.loaders': ['pyscript = '
                         'kraken.core.loaders.pyscript:PyscriptLoader']}

setup_kwargs = {
    'name': 'kraken-core',
    'version': '0.1.1',
    'description': '',
    'long_description': '# kraken-core\n\nThe kraken core API provides the primitives to fully describe complex build processes of software components.\n\n## 1. Concepts\n\n### 1.1 Projects\n\nA project maps to a directory on the file system that represents a software component, usually\ncomprised of many different source code and supporting files, and eventually additional other\nsub projects.\n\nIn every build there is at least one project involved, the root project. When tasks are referenced\nby the full qualified name, the root project name is omitted from the path (similar to how the root\ndirectory in a file system does not have a name and is represented by a single slash).\n\n### 1.2 Tasks\n\nA task is a logical unit of work that has a unique name within the project it is associated with. Tasks\ncan have dependencies on other tasks or groups of tasks, even across projects. When a task is optional,\nit will only be executed if strictly required by another task.\n\nDependencies on task can be strict or non-strict. Non-strict dependencies enforce an order on the\nexecution sequence of tasks, wheras strict dependencies will ensure that the dependency has been\nexecuted.\n\nTasks are usually marked as "default", meaning that they are selected by default if no task selectors are\nspecified (see below on [Task selectors](#14-task-selectors)).\n\nEvery task is associated with one [Action](#15-actions) that is executed for the task.\n\n### 1.4 Task selectors\n\nWithout any arguments, all default tasks are selected and executed. When an explicit selection is made,\nit can be in one of the following forms:\n\n1. A fully qualified project reference\n2. A fully qualified task reference\n3. A task name to select from all projects that contain it\n\nProjects and tasks are structured hierarchally and fully qualified references are constructed like file system\npaths but with colons instead of slashes. For example, `:` represents the root roject, `:foo:bar` references\ntask or project `bar` in project `foo` in the root project, `spam` references all tasks named `spam`.\n\n### 1.5 Actions\n\nAn action is a unit of work that is executed by a task. A task\'s action may be set on creation or only right before\nit needs to be executed, in the `Task.finalize()` method.\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
